#ifndef _ACQUIRESOUND_HPP_
#define _ACQUIRESOUND_HPP_

/*
struct DATA
{
	EST_SingleChannelWave wave;
	// other stuff
	void process() // initialize other stuff from wave
	{
	}
}


template <class DATA>
class AcquireSound
{
	AcquireSound(int nChunks, real period, int freq, int depth);
	~AcquireSound() ;
	void start();
	void waitUpdated();
	DATA& getWave(int i);
};


*/

#include "cyril-macro.h"
#include "cyril-timer.h"

#include "EST.h"
#include "EST_types.h"
#include "EST_error.h"
#include "EST_SingleChannelWave.h"

#include <pthread.h>
#include <semaphore.h>


class AcquireChunk
{
 public:
	EST_SingleChannelWave wave;
	sem_t sem_available;
	pthread_t id;

	real period;
	int freq;
	int depth;
	
	AcquireChunk(real period, int freq, int depth): period(period), freq(freq), depth(depth)
	{
		ASSERTC(sem_init(&sem_available , false, 0),==,0); // 0 token because there is nothing available at the beginning
	}
	~AcquireChunk()
	{
		ASSERTC(pthread_cancel(id),==,0);
		ASSERTC(sem_destroy(&sem_available),==,0);
	}	
	void start()
	{
		ASSERTC(pthread_create(&id, NULL, (threadfn)&start_, (void*)this),==,0);
	}

	
	static void start_(AcquireChunk *data)
	{
		EST_Option al;
		al.add_item("-sample_rate", xtostr(data->freq).c_str());
		al.add_item("-audiodevice", "/dev/dsp");
		al.add_item("-obo", "native");
		al.add_item("-time", xtostr(data->period).c_str());
		al.add_item("-ostype", (data->depth == 1 || data->depth == 8 ? "byte" : "short"));

		while(true)
		{
			ASSERTC(record_wave(data->wave,al),==,0);
			ASSERTC(sem_post(&data->sem_available),==,0);
		}
	}
};




template <class DATA>
class AcquireSound
{
	DATA *waves;
	DATA empty_wave;
	int start_pos, length, size; // start_pos is the index of the first wave, then % size for length elements (usually length=size-1)
	int start_using, length_using;
	sem_t sem_accessing; // protection semaphore for start_pos/length/size variables
	sem_t sem_updated;
	pthread_t id;
	AcquireChunk *ac;

	real period;
	int freq;
	int depth;
	
 	static void start_(AcquireSound *data)
	{
		int res, val_sem;
		Date time;
		data->ac = new AcquireChunk(data->period, data->freq, data->depth);
		data->ac->start();

		while(true)
		{
			// waiting new chunk
			while ((res = sem_wait(&data->ac->sem_available)) == -1); ASSERTC(res,==,0);
			time = get_c_us();
			// copy the chunk in the global array
			int current = (data->length < data->size-1 ? data->start_pos+data->length : data->start_pos+data->size-1) % data->size;
			data->waves[current].wave = data->ac->wave;
			data->waves[current].time = time/1.e-6 - data->ac->wave.num_samples()/(real)data->ac->wave.sample_rate();
			// process the chunk
			data->waves[current].process();
			// and tell that there are new data
			while ((res = sem_wait(&data->sem_accessing)) == -1); ASSERTC(res,==,0);
			ASSERTC(sem_getvalue(&data->sem_updated, &val_sem),==,0);
			if (data->length < data->size-1) 
				data->length++; else
				data->start_pos = (data->start_pos+1)%data->size;
			if (val_sem == 0)
				ASSERTC(sem_post(&data->sem_updated),==,0);
			ASSERTC(sem_post(&data->sem_accessing),==,0);
		}
	}
	
public:
	
	AcquireSound(int nChunks, real period, int freq, int depth): period(period), freq(freq), depth(depth)
	{
		start_pos = start_using = 0;
		length = length_using = 0; // for now, then it will increase up to n_elements
		size = nChunks+1;
		waves = new DATA[size];
		ASSERTC(sem_init(&sem_accessing , false, 1),==,0); // 1 token because data are not accessed by default
		ASSERTC(sem_init(&sem_updated , false, 0),==,0); // 0 token because not updated at the beginning
	}
	~AcquireSound() 
	{ 
		delete[] waves;
		delete ac;
		ASSERTC(pthread_cancel(id),==,0);
		ASSERTC(sem_destroy(&sem_accessing),==,0);
		ASSERTC(sem_destroy(&sem_updated),==,0);
	}
	void start()
	{
		ASSERTC(pthread_create(&id, NULL, (threadfn)&start_, (void*)this),==,0);
	}


	void waitUpdated()
	{
		int res;
		// wait updated
		while ((res = sem_wait(&sem_updated)) == -1); ASSERTC(res,==,0);
		// get access to data
		while ((res = sem_wait(&sem_accessing)) == -1); ASSERTC(res,==,0);
		// get data
		start_using = start_pos;
		length_using = length;
		// free data mutex
		ASSERTC(sem_post(&sem_accessing),==,0);
	}
	
	/*
	return the chunk, or a wave with length 0
	*/
	DATA& getWave(int i)
	{
		if (i < 0 || i >= length_using) return empty_wave;
		return waves[(start_using+i)%size];
	}
};




#endif

